{% extends "_base.html" %}
{% block title %}Authentication{% endblock %}

{% block body %}
<form method="post" action="{{ myurl }}">
    <div>
        {% if error is defined %}
            <p class="error"><strong>Error: {{ error }}</strong></p>
        {% endif %}
        <p>
           Please enter a registration token.
        </p>
        <input type="hidden" name="session" value="{{ session }}" />
        <input type="text" name="token" />
        <input type="submit" value="Authenticate" />
    </div>
</form>
{% endblock %}
